-------------------------------------------------------------------------------
-- Concurrent Haskell Debugger 
--   ConcurrentMain Module
--     by Thomas Boettcher <thomas.boettcher@gmx.de>
-------------------------------------------------------------------------------

-- the main interface is completed, some functions needed. see below!

module CHD.ConcurrentMain
  ( 
    C.ThreadId,
    forkIO,		-- :: IO () -> IO ThreadId
    forkIOLine,
    C.myThreadId,	-- :: IO ThreadId
    killThread,		-- :: ThreadId -> IO ()
    killThreadLine,	-- :: CodePosition -> ThreadId -> IO ()
    markThread,		-- :: ThreadId -> IO ()
    unmarkThread,	-- :: ThreadId -> IO ()
    --par,		-- :: a -> b -> b
    --seq,		-- :: a -> b -> b
    --fork,		-- :: a -> b -> b
    yield,		-- :: IO ()

    threadDelay,	-- :: Int -> IO ()
    threadWaitRead,	-- :: Int -> IO ()
    threadWaitWrite,	-- :: Int -> IO ()

    mergeIO,		-- :: [a]   -> [a] -> IO [a]
    nmergeIO		-- :: [[a]] ->y IO [a]
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

-- the following is as described in the diploma thesis

import qualified Control.Concurrent as C

import CHD.DebugInterface
import CHD.DebugMsgChan
import CHD.DebugMain
import CHD.GuiMain
import CHD.GuiMsgChan
import CHD.BaseTypes


----------------------------------------------
-- not yet implemented:

import Control.Concurrent
  (
    --par,		-- :: a -> b -> b
    --seq,		-- :: a -> b -> b
    --fork,		-- :: a -> b -> b
    yield,		-- :: IO ()

    threadWaitRead,	-- :: Int -> IO ()
    threadWaitWrite,	-- :: Int -> IO ()

    mergeIO,		-- :: [a]   -> [a] -> IO [a]
    nmergeIO		-- :: [[a]] ->y IO [a]  
  )


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

forkIO :: IO () -> IO C.ThreadId
forkIO = forkIOLine NoPosition


forkIOLine :: CodePosition -> IO () -> IO C.ThreadId
forkIOLine pos thread = do
  forkStop <- C.newEmptyMVar
  newId <- C.forkIO (do
      C.takeMVar forkStop
      debugStop1 <- sendDebugMsg ThreadStart
      C.takeMVar debugStop1
      thread
      sendDebugMsg ThreadEnd
      return ()
    )
  debugStop2 <- sendDebugMsg (ThreadForkSuspend (Identity newId) pos)
  C.takeMVar debugStop2
  debugStop3 <- sendDebugMsg (ThreadFork (Identity newId) (CHD forkStop) pos)
  C.takeMVar debugStop3
  return newId


killThread :: C.ThreadId -> IO ()
killThread = killThreadLine NoPosition


killThreadLine :: CodePosition -> C.ThreadId -> IO ()
killThreadLine pos threadId = do
  debugStop1 <- sendDebugMsg (ThreadKillSuspend (Identity threadId) pos)
  C.takeMVar debugStop1
  C.killThread threadId
  debugStop2 <- sendDebugMsg (ThreadKill (Identity threadId) pos)
  C.takeMVar debugStop2


threadDelay :: Int -> IO ()
threadDelay time = do
  debugStop1 <- sendDebugMsg (ThreadDelaySuspend)
  C.takeMVar debugStop1
  C.threadDelay time
  debugStop2 <- sendDebugMsg (ThreadDelay)
  C.takeMVar debugStop2


markThread :: String -> C.ThreadId -> IO ()
markThread color threadId = do
  sendDebugMsg (ThreadMark (Identity threadId) color)
  return ()


unmarkThread :: C.ThreadId -> IO ()
unmarkThread threadId = do
  sendDebugMsg (ThreadUnmark (Identity threadId))
  return ()
