-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   BaseFunctions Module
--
-------------------------------------------------------------------------------

module CHS.BaseFunctions
  (
   dummyId,
   without,
   deleteWithPos,
   deleteWithKey,
   replaceWithKey,
   replaceWithPos,
   showObjectList,
   objectList,
   nextObject,
   dependentOn
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO, print, putStr, getChar, getLine )
import qualified Prelude as P
import qualified Control.Concurrent as C
import System.IO.Unsafe

import CHS.BaseTypes


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

dummyId :: C.ThreadId
dummyId = unsafePerformIO (C.forkIO (return ()))


without :: Eq a => [a] -> [a] -> [a]
without list1 list2 =
  filter (\x -> not $ elem x list2) list1


deleteWithPos :: Int -> [a] -> [a]
deleteWithPos _ [] = error "nicht in Liste enthalten"
deleteWithPos 1 (_:xs) = xs
deleteWithPos (n+1) (x:xs) = x:deleteWithPos n xs


deleteWithKey :: Eq a => a -> [(a,b)] -> (Maybe Int,[(a,b)])
deleteWithKey _ [] = (Nothing,[])
deleteWithKey y ((x1,x2):xs) 
  | x1 == y   = (Just 1,xs)
  | otherwise = let (maybeV,list) = deleteWithKey y xs in
		(fmap (+1) maybeV,(x1,x2):list)


replaceWithKey :: Eq a => (a,b) -> [(a,b)] -> [(a,b)]
replaceWithKey y [] = [y]
replaceWithKey (y1,y2) ((x1,_):xs) | x1 == y1 = ((y1,y2):xs)
replaceWithKey y (x:xs) = x:replaceWithKey y xs


replaceWithPos :: Int -> (a,b) -> [(a,b)] -> [(a,b)]
replaceWithPos _ _ [] = []
replaceWithPos 1 y (_:xs) = (y:xs)
replaceWithPos (n+1) y (x:xs) = x:replaceWithPos n y xs


objectList :: [Thread] -> [Either Object C.ThreadId]
objectList = map (nextObject . snd)
   

nextObject :: IO () -> Either Object C.ThreadId
nextObject (Return _) = Left NoObject

nextObject (ConcAction obj _ _ _) = Left obj

nextObject (KillAction tId _ _) = Right tId

nextObject (ForkAction _ _ _) = Left NoObject

nextObject (SeqAction _ _) = Left NoObject

nextObject (Bind (Return _) _) = Left NoObject

nextObject (Bind (ConcAction obj _ _ _) _) = Left obj

nextObject (Bind (ForkAction _ _ _) _) = Left NoObject

nextObject (Bind (KillAction tId _ _) _) = Right tId

nextObject (Bind (SeqAction _ _) _) = Left NoObject

nextObject (Bind (Bind fa1 fa2) fa3) = 
  nextObject (Bind fa1 (\x->(Bind (fa2 x) fa3)))


dependentOn :: Int -> Either Object C.ThreadId -> [Thread] -> [Int]
dependentOn _ _ [] = []

dependentOn _ (Left NoObject) _ = []

dependentOn pos obj ((_,io):ts)
  | obj==nextObject io = pos:dependentOn (pos+1) obj ts
  | otherwise	       = dependentOn (pos+1) obj ts


showObjectList :: [Either Object C.ThreadId] -> String
showObjectList [] = "[]"

showObjectList [Left (MVarObj (MVarNo mvarNo))] =
  show mvarNo	

showObjectList [Left (SampleVarObj (SampleVarNo samplevarNo))] =
  show samplevarNo

showObjectList [Left (QSemObj (QSemNo qsemNo))] =
  show qsemNo	

showObjectList [Left (QSemNObj (QSemNNo qsemnNo))] =
  show qsemnNo	

showObjectList [Left (ChanObj (ChanNo chanNo))] =
  show chanNo	

showObjectList [Left NoObject] =
  "-1"

showObjectList (Right _:qs) =
  " Thread "

showObjectList (Left (ThreadObj (ThreadNo threadNo)):ts) =
  show threadNo ++ "  " ++ showObjectList ts

showObjectList (Left (MVarObj (MVarNo mvarNo)):ms) =
  show mvarNo ++ "  " ++ showObjectList ms

showObjectList (Left (SampleVarObj (SampleVarNo samplevarNo)):ms) =
  show samplevarNo ++ "  " ++ showObjectList ms

showObjectList (Left (QSemObj (QSemNo qsemNo)):qs) =
  show qsemNo ++ "  " ++ showObjectList qs

showObjectList (Left (QSemNObj (QSemNNo qsemnNo)):qs) =
  show qsemnNo ++ "  " ++ showObjectList qs

showObjectList (Left (ChanObj (ChanNo chanNo)):qs) =
  show chanNo ++ "  " ++ showObjectList qs

showObjectList (Left NoObject:ms) =
  "-1" ++ "  " ++ showObjectList ms

showObjectList (Right _:qs) =
  " Thread " ++ showObjectList qs


