-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   CheckMain Module
--   
-------------------------------------------------------------------------------

module CHS.CheckMain 
  (
    breakMVar,
    finishedBreakMVar,
    loopCheck
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CHD

import CHS.BaseTypes
import CHS.BaseFunctions
import CHS.ComMVars
import Tree


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

minimized = True


startDepth :: Int
startDepth = 4


endDepth :: Int
endDepth = 20


stepDepth :: Int -> Int
stepDepth depth =
  let step = div 40 depth in
  if step > 0
     then step
     else 1


loopCheck :: [Thread] -> Maybe [C.ThreadId] -> P.IO ()
loopCheck ts oldRes = do
  (checkRes,tree) <- case oldRes of
       Nothing         -> check ts startDepth
       Just path       -> return (Just path,Node [] [])
  --writeFile "./graph.dot" (toDottyFile tree)
  case checkRes of
       Nothing           -> putStrLn "Nothing"
       Just (tId:x)      -> CHD.markThread "black" tId
       Just []	         -> putStrLn "DeadLock"
  C.putMVar finishedBreakMVar ()
  C.takeMVar breakMVar
  result <- C.takeMVar treeMVar
  case result of
       Stepped thread@(threadId,_) -> 
	       let ts' = replaceWithKey thread ts in
	       case checkRes of
		    Nothing	   -> loopCheck ts' Nothing
		    Just (tId:ids) -> do
				      CHD.unmarkThread tId
				      if threadId==tId
					 then loopCheck ts' (Just ids)
					 else loopCheck ts' Nothing
		    _	           -> error "Error in loopcheck"
       Fork b thread -> 
	      let ts' = replaceWithKey b (replaceWithKey thread ts) in
	      case checkRes of 
		   Nothing      -> loopCheck ts' Nothing
		   Just (tId:_) -> CHD.unmarkThread tId >>
				   loopCheck ts' Nothing
		   _		-> error "Error in loopcheck"
       Kill killId thread -> 
	      let (_,ts') = deleteWithKey killId (replaceWithKey thread ts) in
	      case checkRes of 
		   Nothing      -> loopCheck ts' Nothing
		   Just (tId:_) -> CHD.unmarkThread tId >>
				   loopCheck ts' Nothing
		   _		-> error "Error in loopcheck"
       _ -> error "Error in loopcheck"


check :: [Thread] -> Int -> P.IO (Maybe [C.ThreadId],Tree)
check ts depth = do
  (chRes,tree) <- checkThreads ts depth True [1..length ts] []
  putStrLn $ show depth
  b <- C.isEmptyMVar breakMVar
  if b && depth <= endDepth
     then case chRes of
	       Nothing   -> check ts (depth+(stepDepth depth))
	       Just path -> return (Just path,tree)
     else return (chRes,tree)


checkThreads :: [Thread] -> Int -> Bool -> [Int] -> [Int]
			 -> P.IO (Maybe [C.ThreadId],Tree)
checkThreads [] _ _ _ _ =
  return (Just [],Node [] [])

checkThreads ts@(_:_) 0 _ _ _ =
  return (Nothing,Node [Label (showObjectList (objectList ts))] [])

checkThreads ts@(_:_) (n1@(n+1)) dead (m:ms) searchList = do
  b <- C.isEmptyMVar breakMVar
  if b
     then do
       let objects = Label (showObjectList (objectList ts))
	   thread = ts!!(m-1)
	   threadId = fst thread
	   dependList1 = dependentOn 1 (nextObject (snd thread)) ts
	   dependList2 = dependentOn 1 (Right threadId) ts
	   dependList = dependList1 ++ dependList2
	   list = filter (>m) dependList
	   edge = Label (show (m,list))
       resultT' <- checkThread thread
       case resultT' of
	 NotStepped (t',restoreAction) -> do
	   let ts' = replaceWithPos m t' ts
	   (checkRes,tree) <- checkThreads ts' n1 dead (m:ms) searchList
	   restoreAction
	   return (checkRes,Node [objects] [Edge [edge,Style Dashed] tree])
	 Suspended -> do
	   checkThreads ts n1 dead ms searchList
	 Stop -> do
	   checkThreads ts n1 False ms searchList
	 Terminate -> do
	   let ts' = deleteWithPos m ts
	   (checkRes,tree) <- 
	     if minimized
		then checkThreads ts' n True (list++[1..m-1]) 
				  ([m..length ts'] `without` list)
		else checkThreads ts' n True [1..(length ts')-1] []
	   case checkRes of
	     Nothing -> do
	       (checkRes',tree') <- checkThreads ts n1 False ms searchList
	       return (checkRes',addSuccessor tree' [] [edge] tree)
	     Just path ->
	       return (Just (threadId:path),
		       Node [objects,Color Red] [Edge [edge,Color Red] tree])
	 Stepped (t',restoreAction) -> do
	   let ts' = replaceWithPos m t' ts
	   (checkRes,tree) <- 
	     if minimized
		then checkThreads ts' n True (list++[1..m])
				  ([m+1..length ts'] `without` list)
		else checkThreads ts' n True [1..length ts'] []
	   restoreAction
	   case checkRes of
	     Nothing -> do
	       (checkRes',tree') <- checkThreads ts n1 False ms searchList
	       return (checkRes',addSuccessor tree' [] [edge] tree)
	     Just path -> do
	       (checkRes',tree') <-
	         checkThreads ts (length path) False ms searchList
	       case checkRes' of
		 Nothing -> do 
		  return $ 
		   (Just (threadId:path),
		    (addSuccessor tree' [Color Red] [edge,Color Red] tree))
		 Just path' -> return $ 
		   (Just path',
		    (addSuccessor tree' [Color Red] [edge,Color Red] tree))
         Fork newT (t',restoreAction) -> do
	   let ts' = replaceWithPos m t' ts ++ [newT]
	   (checkRes,tree) <- 
	     {-if minimized
		then checkThreads ts' n True (list++(length ts':[1..m]))
				  ([m+1..(length ts')-1] `without`list)-}
		checkThreads ts' n True [1..length ts'] []
	   restoreAction
	   case checkRes of
	     Nothing -> do
	       (checkRes',tree') <- checkThreads ts n1 False ms searchList
	       return (checkRes',addSuccessor tree' [] [edge] tree)
	     Just path -> return
	       (Just (threadId:path),Node [objects,Color Red] 
					  [Edge [edge,Color Red] tree])
         Kill killId (t',restoreAction) -> do
	   let (maybeV,ts') = deleteWithKey killId (replaceWithPos m t' ts)
	   (checkRes,tree) <- case maybeV of
	     Nothing -> 
	       if minimized
		  then checkThreads ts' n True (list++[1..m]) 
				    (([m+1..(length ts')] `without` list))
		  else checkThreads ts' n True [1..length ts'] [] 
	     Just del ->
	       let greater = map (\i -> i-1) $ filter (>del) (list++[1..m])
		   less = filter (<del) (list++[1..m])
		   newList = greater++less in
	       if minimized
		  then checkThreads ts' n True newList 
				    (([1..length ts'] `without` newList))
		  else checkThreads ts' n True [1..length ts'] []
	   restoreAction
	   case checkRes of
	     Nothing -> do
	       (checkRes',tree') <- checkThreads ts n1 False ms searchList
	       return (checkRes',addSuccessor tree' [] [edge] tree)
	     Just path -> 
		return (Just (threadId:path),
			Node [objects,Color Red] [Edge [edge,Color Red] tree])

     else return (Nothing,Node [Label (showObjectList (objectList ts))] [])

checkThreads ts@(_:_) _ dead [] searchList = do
  let objects = Label (showObjectList (objectList ts))
  b <- C.isEmptyMVar breakMVar
  if b 
     then if dead
	     then do
		  allDead <- checkSuspended ts searchList
		  if allDead
		     then return (Just [],Node [objects,Color Red] 
				 (map (\m -> Edge [Label (show m),Color Red]
						  (Node [Label "",Color Red] 
							[])) 
				      searchList))
		     else return (Nothing,Node [objects]
				 (map (\m -> Edge [Label (show m),Color Green]
						  (Node [Label "",Color Green]
							[])) 
				      searchList))
	     else return (Nothing,Node [objects] [])
     else return (Nothing,Node [objects] [])

checkThreads (_:_) n b _ _ =
  error (show (n,b))


checkThread :: Thread -> P.IO (Result (Thread,P.IO ()))
checkThread (_,(Return _)) =
  return Terminate

checkThread (id,(KillAction killId _ kA)) = do
  v <- kA
  return $ Kill killId ((id,Return v),return ())

checkThread (id,(ForkAction b _ cA)) = do
  (v,tId) <- cA
  return $ Fork (tId,b) ((id,Return v),return ())

checkThread (id,(ConcAction _ _ cA restoreActionF)) = do
  maybeV <- cA
  case maybeV of
       Nothing  -> return Suspended
       Just (v1,v2) -> return $ Stepped ((id,Return v1),restoreActionF v2)

checkThread (id,(SeqAction (Ignorable sA) _)) = do
  v <- sA
  return $ NotStepped ((id,Return v),return ())

checkThread (id,(SeqAction Executable sA)) = do
  v <- sA
  return $ NotStepped ((id,Return v),return ())

checkThread (id,(SeqAction NonBackTrackable _)) =
  return Stop

checkThread (id,(Bind (ConcAction _ _ cA restoreActionF) fa)) = do
  maybeV <- cA
  case maybeV of
       Nothing -> return Suspended
       Just (v1,v2) -> return $ Stepped ((id,fa v1),restoreActionF v2)

checkThread (id,(Bind (SeqAction (Ignorable sA) _) fa)) = do
  v <- sA
  return $ NotStepped ((id,fa v),return ())

checkThread (id,(Bind (SeqAction Executable sA) fa)) = do
  v <- sA
  return $ NotStepped ((id,fa v),return ())

checkThread (id,(Bind (SeqAction NonBackTrackable _) _)) =
  return Stop

checkThread (tId,(Bind (SeqAction (MyThreadId fId) _) fa)) = do
  v <- fId tId
  return $ NotStepped ((tId,fa v),return ())

checkThread (id,(Bind (Return a) fa)) =
  return $ NotStepped ((id,fa a),return ())

checkThread (id,(Bind (ForkAction b _ cA) fa)) = do
  (v,newId) <- cA
  return $ Fork (newId,b) ((id,fa v),return ())

checkThread (id,(Bind (KillAction killId _ kA) fa)) = do
  v <- kA
  return $ Kill killId ((id,fa v),return ())

checkThread (tId,(Bind (Bind fa1 fa2) fa3)) = do
  checkThread (tId,(Bind fa1 (\x->(Bind (fa2 x) fa3))))


checkSuspended :: [Thread] -> [Int] -> P.IO Bool
checkSuspended _ [] = return True

checkSuspended ts (m:ms) = do
   let thread = ts!!(m-1)
   resultT' <- checkThread thread
   case resultT' of
	Stepped (_,restoreAction) -> restoreAction >>
				     return False
	Fork _ (_,restoreAction)  -> restoreAction >>
				     return False
	Kill _ (_,restoreAction)  -> restoreAction >>
				     return False
	Stop -> return False
	NotStepped (t',_) -> 
	  let ts' = replaceWithPos m t' ts in
	  checkSuspended ts' ms
	Terminate -> checkSuspended ts ms
	Suspended -> checkSuspended ts ms