-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperChan Module
--   
-------------------------------------------------------------------------------

module CHS.StepperChan
  (
    Chan(..),
    newChan,         -- :: IO (Chan a)
--    newChanLine,     -- :: CodePosition -> IO (Chan a)
    writeChan,       -- :: Chan a -> a -> IO ()
    writeChanLine,   --	:: CodePosition -> Chan a -> a -> IO ()
--    writeChanLabel,  -- :: Chan a -> a -> String -> IO ()
    readChan,        -- :: Chan a -> IO a
    readChanLine,    -- :: CodePosition -> Chan a -> IO a
--    dupChan,         -- :: Chan a -> IO (Chan a)
--    unGetChan,       -- :: Chan a -> a -> IO ()
--    getChanContents, -- :: Chan a -> IO [a]
--    writeList2Chan,  -- :: Chan a -> [a] -> IO ()
    isEmptyChan     --	:: Chan a -> Bool
--    isEmptyChanLine  --	:: CodePosition -> Chan a -> Bool
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CD

import CHS.BaseTypes
import CHS.ComMVars

import System.IO.Unsafe


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data Chan a = Chan (CD.Chan a) (C.Chan a) (C.Chan ())


-------------------------------------------------------------------------------
-- Chan-Aktionen
-------------------------------------------------------------------------------

newChanNoMVar :: C.MVar ChanNo
newChanNoMVar = unsafePerformIO $ C.newMVar (ChanNo (-1))


newChan :: IO (Chan a)
newChan =
  ConcAction (ChanObj (ChanNo (-1))) 
	     (do cdChan <- CD.newChan
	         C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 searchChan <- C.newChan
		 readStop <- C.newChan
		 return $ Chan cdChan searchChan readStop)
	     (do cChan <- C.newChan
		 searchChan <- C.newChan
		 readStop <- C.newChan
		 chanNo <- C.takeMVar newChanNoMVar
		 let cdChan = CD.Chan chanNo cChan
		     chan = Chan cdChan searchChan readStop
		 C.putMVar newChanNoMVar (pred chanNo)    
		 return $ Just (chan,chan))
             (\_ -> do chanNo <- C.takeMVar newChanNoMVar
		       C.putMVar newChanNoMVar (succ chanNo))


writeChan :: Chan a -> a -> IO ()
writeChan = writeChanLine NoPosition


writeChanLine :: CodePosition -> Chan a -> a -> IO ()
writeChanLine pos (Chan cdChan@(CD.Chan chanNo _) searchChan readStop) value = 
  ConcAction (ChanObj chanNo)
	     (do CD.writeChanLine pos cdChan value
		 C.putMVar breakMVar ()
		 C.writeChan readStop ()
		 C.takeMVar finishedBreakMVar
		 C.writeChan searchChan value)
	     (do b <- C.isEmptyChan searchChan
		 if b
		    then do C.writeChan searchChan value
			    return $ Just ((),())
		    else return Nothing)
             (\_ -> do C.readChan searchChan
		       return ())


readChan :: Chan a -> IO a
readChan = readChanLine NoPosition


readChanLine :: CodePosition -> Chan a -> IO a
readChanLine pos (Chan cdChan@(CD.Chan chanNo _) searchChan readStop) =
  ConcAction (ChanObj chanNo)
	     (do CD.readChanLine pos cdChan
		 C.readChan readStop
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 C.readChan searchChan)
	     (do b <- C.isEmptyChan searchChan
                 if b
                    then return Nothing
                    else do v <- C.readChan searchChan
			    return $ Just (v,v))
             (\v -> C.writeChan searchChan v)


--dupChan :: Chan a -> IO (Chan a)

--unGetChan :: Chan a -> a -> IO ()


isEmptyChan :: Chan a -> IO Bool
isEmptyChan (Chan cdChan@(CD.Chan chanNo _) searchChan _) = 
  ConcAction (ChanObj chanNo)
	     (do CD.isEmptyChan cdChan
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 C.isEmptyChan searchChan)
	     (do b <- C.isEmptyChan searchChan
		 return $ Just (b,b))
             (\_ -> return ())


--getChanContents :: Chan a -> IO [a]

--writeList2Chan :: Chan a -> [a] -> IO ()