-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperMain Module
--   
-------------------------------------------------------------------------------

module CHS.StepperMain
  (
  forkIO,		-- :: IO () -> IO C.ThreadId
  forkIOLine,		-- :: CodePosition -> IO () -> IO C.ThreadId
  forkIOLabel,		-- :: String -> IO () -> IO C.ThreadId
  forkIOLabelLine,	-- :: CodePosition -> String -> IO () -> IO C.ThreadId
  killThread,		-- :: C.ThreadId -> IO ()
  killThreadLine,	-- :: CodePosition -> C.ThreadId -> IO ()
  threadDelay,		-- :: Int -> IO ()
  myThreadId,		-- :: IO ThreadId
  C.ThreadId
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CD

import CHS.StepMain
import CHS.BaseTypes
import CHS.BaseFunctions
import CHS.ComMVars


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

forkIO :: IO () -> IO C.ThreadId
forkIO = forkIOLine NoPosition 


forkIOLine :: CodePosition -> IO () -> IO C.ThreadId
forkIOLine pos a = 
  ForkAction a
	     (do threadId <- CD.forkIOLine pos (executeIO a)
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 return (threadId,threadId))
	     (do threadId <- C.forkIO (return ())
		 return (threadId,threadId))


forkIOLabel :: String -> IO () -> IO C.ThreadId
forkIOLabel label a = 
  ForkAction a
	     (do threadId <- CD.forkIOLabel label (executeIO a)
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 return (threadId,threadId)) 
	     (do threadId <- C.forkIO (return ())
	         return (threadId,threadId))


forkIOLabelLine :: CodePosition -> String -> IO () -> IO C.ThreadId
forkIOLabelLine pos label a = 
  ForkAction a
	     (do threadId <- CD.forkIOLabelLine pos label (executeIO a)
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 return (threadId,threadId)) 
	     (do threadId <- C.forkIO (return ())
	         return (threadId,threadId))


killThread :: C.ThreadId -> IO ()
killThread = killThreadLine NoPosition


killThreadLine :: CodePosition -> C.ThreadId -> IO ()
killThreadLine pos threadId =
  KillAction threadId 
	     (do CD.killThreadLine pos threadId
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 return ())
	     (return ())


threadDelay :: Int -> IO ()
threadDelay time =
  SeqAction (Ignorable (return ())) (CD.threadDelay time)


myThreadId :: IO C.ThreadId
myThreadId =
  SeqAction (MyThreadId (\tId -> return tId)) (CD.myThreadId)