-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperQSemN Module
--   
-------------------------------------------------------------------------------

module CHS.StepperQSemN
  (
   QSemN,
   newQSemN,		-- :: Int -> IO QSemN
   newQSemNLine,	-- :: CodePosition -> Int -> IO QSemN
   signalQSemN,		-- :: Int -> QSemN -> IO ()
   signalQSemNLine,	-- :: CodePosition -> Int -> QSemN -> IO ()
   waitQSemN,		-- :: Int -> QSemN -> IO ()
   waitQSemNLine	-- :: CodePosition -> Int -> QSemN -> IO ()   
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CD
import qualified System.IO as SI

import CHS.BaseTypes
import CHS.ComMVars

import System.IO.Unsafe

import qualified QSemN as QN


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data QSemN = QSemN CD.QSemN QN.QSemN QN.QSemN


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- QSem-Aktionen
-------------------------------------------------------------------------------

newQSemNNoMVar :: C.MVar QSemNNo
newQSemNNoMVar = unsafePerformIO $ C.newMVar (QSemNNo (-1))


newQSemN :: Int -> IO QSemN
newQSemN x = newQSemNLine NoPosition x


newQSemNLine :: CodePosition -> Int -> IO QSemN
newQSemNLine pos x = 
  ConcAction (QSemNObj (QSemNNo (-1))) 
	     (do cdQSemN <- CD.newQSemNLine pos x
	         C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 searchQSemN <- QN.newQSemN x
		 waitStop <- QN.newQSemN x
		 return $ QSemN cdQSemN searchQSemN waitStop)
	     (do qsemn <- C.newQSemN x
		 searchQSemN2 <- QN.newQSemN x
		 waitStop2 <- QN.newQSemN x
		 qsemnNo <- C.takeMVar newQSemNNoMVar
		 let cdQSemN2 = CD.QSemN qsemnNo qsemn
		     qsemn2 = QSemN cdQSemN2 searchQSemN2 waitStop2
		 C.putMVar newQSemNNoMVar (pred qsemnNo)
		 return $ Just (qsemn2,qsemn2))
             (\_ -> do qsemnNo <- C.takeMVar newQSemNNoMVar
		       C.putMVar newQSemNNoMVar (succ qsemnNo))


signalQSemN :: QSemN -> Int -> IO ()
signalQSemN = signalQSemNLine NoPosition 


signalQSemNLine :: CodePosition -> QSemN -> Int -> IO ()
signalQSemNLine pos (QSemN cdQSemN@(CD.QSemN qsemNo _) searchQSemN waitStop) 
		quantity =
  ConcAction (QSemNObj qsemNo) 
	     (do CD.signalQSemNLine pos cdQSemN quantity 
		 C.putMVar breakMVar ()
		 QN.signalQSemN waitStop quantity 
		 C.takeMVar finishedBreakMVar
		 QN.signalQSemN searchQSemN quantity)
	     (do debug "signal 1"
		 QN.signalQSemN searchQSemN quantity
		 debug "signal 2"
		 return $ Just ((),()))
	     (\_ -> QN.waitQSemN searchQSemN quantity)


waitQSemN :: QSemN -> Int -> IO ()
waitQSemN = waitQSemNLine NoPosition


waitQSemNLine :: CodePosition -> QSemN -> Int -> IO ()
waitQSemNLine pos (QSemN cdQSemN@(CD.QSemN qsemNo _) searchQSemN waitStop) 
	      quantity =
  ConcAction (QSemNObj qsemNo)
	     (do CD.waitQSemNLine pos cdQSemN quantity
		 QN.waitQSemN waitStop quantity
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 QN.waitQSemN searchQSemN quantity)
	     (do debug "wait 1"
		 q <- QN.getQuantityQSemN searchQSemN
		 debug "wait 2"
                 if q>=quantity
                    then do debug "wait 3"
			    QN.waitQSemN searchQSemN quantity
			    debug "wait 4"
			    return $ Just ((),())
		    else return Nothing)
             (\_ -> QN.signalQSemN searchQSemN quantity)


debug _ = return ()